import asyncio
import json
from pathlib import Path
from typing import List, Optional
from urllib.parse import urlencode, urljoin

from dynaconf import settings
from playwright.async_api import Browser, Page, Playwright, PlaywrightContextManager, Response
from playwright_stealth import StealthConfig, stealth_async

from tiktokpy.utils.client import block_resources_and_sentry
from tiktokpy.utils.logger import logger


class Client:
    """Browser client"""

    def __init__(self):
        self.base_url = settings.BASE_URL
        self.playwright: Playwright

        self.cookies = json.loads(settings.get("COOKIES", "[]"))

    async def init_browser(self, headless: bool):
        self.playwright = await PlaywrightContextManager().start()

        params = {
            "headless": headless,
            "args": [
                "--no-sandbox",
                "--disable-setuid-sandbox",
                "--disable-dev-shm-usage",
                "--disable-web-security",
                "--disable-notifications",
            ],
        }

        self.browser: Browser = await self.playwright.chromium.launch(**params)
        self.context = await self.browser.new_context()
        await self.context.add_cookies(self.cookies)
        logger.debug(f"🎉 Browser launched. Options: {params}")

    async def new_page(self, blocked_resources: Optional[List[str]] = None) -> Page:
        page: Page = await self.context.new_page()

        # set stealth mode for tiktok
        await stealth_async(
            page,
            StealthConfig(
                webdriver=True,
                webgl_vendor=True,
                chrome_app=False,
                chrome_csi=False,
                chrome_load_times=False,
                chrome_runtime=False,
                iframe_content_window=True,
                media_codecs=True,
                navigator_hardware_concurrency=4,
                navigator_languages=False,
                navigator_permissions=True,
                navigator_platform=False,
                navigator_plugins=True,
                navigator_user_agent=False,
                navigator_vendor=False,
                outerdimensions=True,
                hairline=False,
            ),
        )

        if blocked_resources is not None:
            await page.route(
                "**/*",
                lambda route: asyncio.create_task(
                    block_resources_and_sentry(route, blocked_resources),
                ),
            )

        return page

    async def goto(
        self,
        url: str,
        page: Page,
        query_params: Optional[dict] = None,
        *args,
        **kwargs,
    ) -> Response:
        full_url = urljoin(self.base_url, url)

        if query_params:
            query_params_ = urlencode(query=query_params)
            full_url = f"{full_url}?{query_params_}"

        return await page.goto(full_url, *args, **kwargs)

    async def screenshot(self, path: str, page: Page):
        Path(path).parent.mkdir(parents=True, exist_ok=True)

        await page.screenshot(path=path)

    @classmethod
    async def create(cls, headless: bool = True):
        self = Client()
        await self.init_browser(headless=headless)

        return self



import asyncio
import logging
import typing
from datetime import datetime
from types import TracebackType
from typing import List, Optional

import humanize
from dynaconf import settings

from tiktokpy.bot.decorators import login_required
from tiktokpy.client import Client
from tiktokpy.client.login import Login
from tiktokpy.client.trending import Trending
from tiktokpy.client.user import User
from tiktokpy.models.feed import FeedItem, FeedItems
from tiktokpy.utils.logger import init_logger, logger
from tiktokpy.utils.settings import load_or_create_settings

from .version import __version__


class TikTokPy:
    def __init__(self, settings_path: Optional[str] = None):
        init_logger(logging.INFO)
        self.started_at = datetime.now()
        self.client: Client
        self.is_logged_in = False

        logger.info("🥳 TikTokPy initialized. Version: {}", __version__)

        load_or_create_settings(path=settings_path)

        if settings.get("COOKIES") and settings.get("USERNAME"):
            logger.info(f"✅ Used cookies of @{settings.USERNAME}")
            self.is_logged_in = True
        else:
            logger.info("🛑 Cookies not found, anonymous mode")

        self.headless: bool = settings.get("HEADLESS", True)
        self.lang: str = settings.get("LANG", "en")

    async def __aenter__(self):
        await self.init_bot()

        return self

    async def __aexit__(
        self,
        exc_type: typing.Type[BaseException] = None,
        exc_value: BaseException = None,
        traceback: TracebackType = None,
    ) -> None:
        logger.debug("🤔Trying to close browser..")

        if exc_type and exc_value:
            logger.debug(f"🐛 Found exception. Type: {exc_type}")

        try:
            await asyncio.wait_for(self.client.browser.close(), timeout=10.0)
            await asyncio.wait_for(self.client.playwright.stop(), timeout=5.0)
        except asyncio.TimeoutError:
            logger.debug("🐛 Timeout reached while closing browser")
        else:
            logger.debug("✋ Browser successfully closed")

        logger.info(
            "✋ TikTokPy finished working. Session lasted: {}",
            humanize.naturaldelta(datetime.now() - self.started_at),
        )

    async def trending(self, amount: int = 50) -> List[FeedItem]:
        logger.info("📈 Getting trending items")

        if amount <= 0:
            logger.warning("⚠️ Wrong amount! Return nothing")
            return []

        items = await Trending(client=self.client).feed(amount=amount)

        logger.info(f"📹 Found {len(items)} videos")
        _trending = FeedItems(__root__=items)

        return _trending.__root__

    @login_required()
    async def follow(self, username: str):
        username = f"@{username.lstrip('@')}"
        await User(client=self.client).follow(username=username)

    @login_required()
    async def like(self, feed_item: FeedItem):
        await User(client=self.client).like(
            username=feed_item.author.username,
            video_id=feed_item.id,
        )

    @login_required()
    async def unlike(self, feed_item: FeedItem):
        await User(client=self.client).unlike(
            username=feed_item.author.username,
            video_id=feed_item.id,
        )

    @login_required()
    async def unfollow(self, username: str):
        username = f"@{username.lstrip('@')}"
        await User(client=self.client).unfollow(username=username)

    async def login_session(self):
        await Login().manual_login()

    async def user_feed(self, username: str, amount: int = 50) -> List[FeedItem]:
        username = f"@{username.lstrip('@')}"
        logger.info(f"📈 Getting {username} feed")
        items = await User(client=self.client).feed(username=username, amount=amount)

        logger.info(f"📹 Found {len(items)} videos")
        feed = FeedItems(__root__=items)

        return feed.__root__

    async def init_bot(self):
        self.client: Client = await Client.create(headless=self.headless)

    @classmethod
    async def create(cls):
        self = TikTokPy()
        await self.init_bot()

        return self

    async def screenshot(self, page, name=""):
        filename = f"{name}_{datetime.now()}".lstrip("_")

        await self.client.screenshot(
            path=f"{settings.HOME_DIR}/screenshots/{filename}.png",
            page=page,
        )



import asyncio
from typing import List

from playwright.async_api import Page, TimeoutError
from tqdm import tqdm

from tiktokpy.client import Client
from tiktokpy.utils import unique_dicts_by_key
from tiktokpy.utils.client import catch_response_and_store, catch_response_info
from tiktokpy.utils.logger import logger

FEED_LIST_ITEM = 'div[data-e2e="recommend-list-item-container"]'
USER_FEED_LIST = 'div[data-e2e="user-post-item-list"]'
USER_FEED_ITEM = f"{USER_FEED_LIST} > div"
USER_FEED_LAST_ITEM = f"{USER_FEED_ITEM}:last-child"
FOLLOW_BUTTON = 'button[data-e2e="follow-button"]'
UNFOLLOW_BUTTON = 'div[class*="DivFollowIcon"]'
MAIN_WRAPPER = "div[class*=DivThreeColumnContainer],main[class*=MainDetailWrapper]"
ERROR_TITLE = "main div[class*=ErrorContainer] p"
SEARCH_USERNAME = 'a[href="/{}"]'


class User:
    def __init__(self, client: Client):
        self.client = client

    async def like(self, username: str, video_id: str):
        page: Page = await self.client.new_page(blocked_resources=["image", "media", "font"])
        logger.debug(f"👥 Like video id {video_id} of @{username}")

        like_info_queue: asyncio.Queue = asyncio.Queue(maxsize=1)

        page.on(
            "response",
            lambda res: asyncio.create_task(
                catch_response_info(res, like_info_queue, "/commit/item/digg"),
            ),
        )

        logger.info(f"🧭 Going to @{username}'s video {video_id} page for like")

        await self.client.goto(
            f"/@{username}/video/{video_id}",
            page=page,
            wait_until="networkidle",
        )

        like_selector = 'span[data-e2e="like-icon"]'
        is_liked = await page.query_selector(f"{like_selector} > div > svg")

        if is_liked:
            logger.info(f"😏 @{username}'s video {video_id} already liked")
            await page.close()

            return

        await page.click(like_selector)

        like_info = await like_info_queue.get()

        if like_info["status_code"] == 0:
            logger.info(f"👍 @{username}'s video {video_id} liked")
        else:
            logger.warning(f"⚠️  @{username}'s video {video_id} probably not liked")

        await page.close()

    async def unlike(self, username: str, video_id: str):
        page: Page = await self.client.new_page(blocked_resources=["image", "media", "font"])
        logger.debug(f"👥 Unlike video id {video_id} of @{username}")

        like_info_queue: asyncio.Queue = asyncio.Queue(maxsize=1)

        page.on(
            "response",
            lambda res: asyncio.create_task(
                catch_response_info(res, like_info_queue, "/commit/item/digg"),
            ),
        )

        logger.info(f"🧭 Going to @{username}'s video {video_id} page for unlike")

        await self.client.goto(
            f"/@{username}/video/{video_id}",
            page=page,
            wait_until="networkidle",
        )

        like_selector = f'{FEED_LIST_ITEM}:first-child span[data-e2e="like-icon"]'
        is_unliked = not await page.query_selector(f"{like_selector} > div > svg")

        if is_unliked:
            logger.info(f"😏 @{username}'s video {video_id} already unliked")
            await page.close()

            return

        await page.click(like_selector)

        unlike_info = await like_info_queue.get()

        if unlike_info["status_code"] == 0:
            logger.info(f"👎 @{username}'s video {video_id} unliked")
        else:
            logger.warning(f"⚠️  @{username}'s video {video_id} probably not unliked")

        await page.close()

    async def follow(self, username: str):
        page: Page = await self.client.new_page(blocked_resources=["image", "media", "font"])
        logger.debug(f"👥 Follow {username}")

        follow_info_queue: asyncio.Queue = asyncio.Queue(maxsize=1)

        page.on(
            "response",
            lambda res: asyncio.create_task(
                catch_response_info(res, follow_info_queue, "/commit/follow/user"),
            ),
        )

        logger.info(f"🧭 Going to {username}'s page for following")

        await self.client.goto(
            f"/@{username.lstrip('@')}",
            page=page,
            wait_until="networkidle",
        )

        follow_title: str = await page.eval_on_selector(
            FOLLOW_BUTTON,
            expression="element => element.textContent",
        )

        if follow_title.lower() != "follow":
            logger.info(f"😏 {username} already followed")
            await page.close()

            return

        await page.click(FOLLOW_BUTTON)

        follow_info = await follow_info_queue.get()

        if follow_info["status_code"] == 0:
            logger.info(f"➕ {username} followed")
        else:
            logger.warning(f"⚠️  {username} probably not followed")

        await page.close()

    async def unfollow(self, username: str):
        page: Page = await self.client.new_page(blocked_resources=["image", "media", "font"])
        logger.debug(f"👥 Unfollow {username}")

        unfollow_info_queue: asyncio.Queue = asyncio.Queue(maxsize=1)

        page.on(
            "response",
            lambda res: asyncio.create_task(
                catch_response_info(res, unfollow_info_queue, "/commit/follow/user"),
            ),
        )

        logger.info(f"🧭 Going to {username}'s page for unfollowing")

        await self.client.goto(
            f"/@{username.lstrip('@')}",
            page=page,
            wait_until="networkidle",
        )

        follow_title: str = await page.eval_on_selector(
            FOLLOW_BUTTON,
            expression="element => element.textContent",
        )

        if follow_title.lower() != "following":
            logger.info(f"😏 {username} already unfollowed")
            return

        await page.click(UNFOLLOW_BUTTON)

        unfollow_info = await unfollow_info_queue.get()

        if unfollow_info["status_code"] == 0:
            logger.info(f"➖ {username} unfollowed")
        else:
            logger.warning(f"⚠️  {username} probably not unfollowed")

        await page.close()

    async def feed(self, username: str, amount: int) -> List[dict]:
        page: Page = await self.client.new_page(blocked_resources=["image", "media", "font"])
        logger.debug(f"📨 Request {username} feed")

        _ = await self.client.goto(
            f"/search/user?q={username.lstrip('@')}",
            page=page,
            wait_until="networkidle",
        )
        username_selector = SEARCH_USERNAME.format(username)

        is_found_user = await page.query_selector(username_selector)

        if not is_found_user:
            logger.error(f'❗️ User "{username}" not found')
            return []

        result: List[dict] = []

        page.on(
            "response",
            lambda res: asyncio.create_task(catch_response_and_store(res, result)),
        )

        try:
            await page.click(username_selector)
            await page.wait_for_selector(MAIN_WRAPPER)
            await page.wait_for_load_state(state="networkidle")
        except TimeoutError:
            logger.error(f'❗️ Unexpected error. Timeout on searching user "{username}"...')
            return []

        logger.debug(f"📭 Got {username} feed")

        error = await page.query_selector(ERROR_TITLE)

        if error:
            logger.info(f'😭 Error message on page: "{await error.text_content()}"')

            return []

        await page.wait_for_selector(USER_FEED_LIST, state="visible")

        await self._paginate_feed_list(
            page=page,
            username=username,
            result=result,
            amount=amount,
        )

        await page.close()
        return unique_dicts_by_key(result, "id")[:amount]

    async def _paginate_feed_list(
        self,
        page: Page,
        username: str,
        result: List[dict],
        amount: int,
    ):
        def result_unique_amount():
            return len(unique_dicts_by_key(result, "id"))

        pbar = tqdm(total=amount, desc=f"📈 Getting {username} feed")
        pbar.n = min(result_unique_amount(), amount)
        pbar.refresh()

        attempts = 0
        max_attempts = 3
        last_result = result_unique_amount()

        is_attempts_limit_reached = attempts >= max_attempts
        is_items_enough = result_unique_amount() < amount

        while is_attempts_limit_reached or is_items_enough:
            logger.debug("🖱 Trying to scroll to last video item")
            await page.evaluate(
                f"""
                document.querySelector('{USER_FEED_LAST_ITEM}')
                    .scrollIntoView();
            """,
            )
            await page.wait_for_timeout(1_000)

            elements = await page.query_selector_all(USER_FEED_ITEM)
            logger.debug(f"🔎 Found {len(elements)} items on page by selector {USER_FEED_ITEM}")

            pbar.n = min(result_unique_amount(), amount)
            pbar.refresh()

            if last_result == result_unique_amount():
                attempts += 1
            else:
                attempts = 0

            if attempts > max_attempts:
                pbar.clear()
                pbar.total = result_unique_amount()
                logger.info(
                    f"⚠️  After {max_attempts} attempts found {result_unique_amount()} videos. "
                    f"Probably some videos are private",
                )
                break

            last_result = result_unique_amount()

            await page.wait_for_timeout(5_000)

        pbar.close()



import json

from dynaconf import loaders, settings

from tiktokpy.client import Client
from tiktokpy.utils.logger import logger
from tiktokpy.utils.settings import BASE_SETTINGS


class Login:
    async def manual_login(self):
        client = await Client.create(headless=False)
        page = await client.new_page()

        await client.goto("/login", page)
        await page.wait_for_selector('div[data-e2e="profile-icon"]', timeout=0)

        username = sub_title = None

        while not all((username, sub_title)):
            await page.hover('div[data-e2e="profile-icon"]')

            await page.wait_for_selector('ul[data-e2e="profile-popup"] > li:first-child')
            # going to "View profile" page
            await page.click('ul[data-e2e="profile-popup"] > li:first-child')

            await page.wait_for_selector('h2[data-e2e="user-title"]', timeout=0)

            username = await page.eval_on_selector(
                'h2[data-e2e="user-title"]',
                expression="element => element.textContent",
            )
            username = username.strip()

            sub_title = await page.eval_on_selector(
                'h1[data-e2e="user-subtitle"]',
                expression="element => element.textContent",
            )

        logger.info(f"🔑 Logged as @{username} aka {sub_title}")

        cookies = await client.context.cookies()

        loaders.write(
            f"{settings.HOME_DIR}/settings.toml",
            {**BASE_SETTINGS, **{"COOKIES": json.dumps(cookies), "USERNAME": username}},
            env="default",
        )

        await client.browser.close()



import asyncio
import random
from typing import List, Optional

from dynaconf import settings
from tqdm import tqdm

from tiktokpy.client import Client
from tiktokpy.utils.client import catch_response_and_store
from tiktokpy.utils.logger import logger

FEED_LIST_ITEM = 'div[data-e2e="recommend-list-item-container"]'
FEED_LIST_ITEM_LAST_CHILD = f"{FEED_LIST_ITEM}:last-child"


class Trending:
    def __init__(self, client: Client, lang: Optional[str] = None):
        self.client = client
        self.lang = lang or settings.get("LANG")

    async def feed(self, amount: int):
        page = await self.client.new_page(blocked_resources=["media", "image", "font"])

        logger.debug('📨 Request "Trending" page')

        result: List[dict] = []

        page.on(
            "response",
            lambda res: asyncio.create_task(catch_response_and_store(res, result)),
        )
        _ = await self.client.goto(
            "/foryou",
            query_params={"lang": self.lang},
            page=page,
            wait_until="networkidle",
        )
        logger.debug('📭 Got response from "Trending" page')

        pbar = tqdm(total=amount, desc=f"📈 Getting trending {self.lang}")
        pbar.n = min(len(result), amount)
        pbar.refresh()

        try:
            while len(result) < amount:
                if len(result) != 0:
                    timeout = random.randint(5, 10)
                    await page.wait_for_timeout(timeout * 1000)

                logger.debug("🖱 Trying to scroll to last video item")

                await page.wait_for_selector(FEED_LIST_ITEM_LAST_CHILD)

                scroll_command = "document.querySelector('{selector}').scrollIntoView();"
                await page.evaluate(scroll_command.format(selector=FEED_LIST_ITEM_LAST_CHILD))
                await page.wait_for_timeout(1_000)

                elements = await page.query_selector_all(FEED_LIST_ITEM)
                logger.debug(f"🔎 Found {len(elements)} items on page")
                await self.client.screenshot("trending.png", page)

                pbar.n = min(len(result), amount)
                pbar.refresh()
        except Exception:
            logger.exception("Something went wrong. Interrupt work")

            # if len(elements) < 500:
            #     logger.debug("🔻 Too less for clearing page")
            #     continue

            # await page.eval_on_selector_all(
            #     f"{FEED_LIST_ITEM_CONTAINER}:not(:last-child)",
            #     expression="(elements) => elements.forEach(el => el.remove())",
            # )
            # logger.debug(f"🎉 Cleaned {len(elements) - 1} items from page")

        await page.close()
        pbar.close()
        return result[:amount]



from functools import wraps
from typing import Any, Callable

from tiktokpy.utils.logger import logger


def login_required(empty_result: Any = None):
    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            async def empty():
                return empty_result

            self_ = args[0]

            if not getattr(self_, "is_logged_in", False):
                logger.error(
                    f'😡 You cannot use function "{func.__name__}" without login! '
                    'Run "tiktokpy login" first and do login',
                )
                return empty()

            return func(*args, **kwargs)

        return wrapper

    return decorator
